'''Parse binary data painless describing the packets' structures in python classes: no need to write 'for' loops or nested 'if' conditionals.'''

__version__ = "0.5.0"

_author  = 'Gehn'
_license = 'GNU LGPLv3'
_url = 'https://github.com/gehnch/bisturi'



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

from bisect import insort, bisect_left, bisect_right
import re, pprint

class Fragments(object):
    def __init__(self, fill=b'.'):
        self.fragments = {}
        self.begin_of_fragments = []
        self.current_offset = 0
        self.fill = fill

    def append(self, string):
        self.insert(self.current_offset, string)

    def extend(self, iterable):
        for string in iterable:
            self.insert(self.current_offset, string)

    def insert(self, position, string):
        #if not string:
        #   return

        #if position in self.fragments:
        #   raise Exception("Collision detected at %08x" % position)

        i = bisect_right(self.begin_of_fragments, position) - 1
        L = len(string)
        if self.fragments:
            b1 = self.begin_of_fragments[i]
            e1 = b1 + len(self.fragments[b1])

            if b1 <= position < e1:
                raise Exception("Collision detected with previous fragment %08x-%08x when inserting new fragment at %08x that span to %08x" % (b1, e1, position, position+L))

            if i+1 < len(self.begin_of_fragments):
                b2 = self.begin_of_fragments[i+1]

                if b2 < position + L:
                    e2 = b2 + len(self.fragments[b2])
                    raise Exception("Collision detected with previous fragment %08x-%08x when inserting new fragment at %08x that span to %08x" % (b2, e2, position, position+L))


        self.begin_of_fragments.insert(i+1, position)

        self.fragments[position] = string
        self.current_offset = position + L

    def tobytes(self):
        begin = 0
        result = []
        for offset, s in sorted(self.fragments.items()):
            result.append(self.fill * (offset-begin))
            result.append(s)
            begin = offset + len(s)

        return b''.join(result)

    def __repr__(self):
        return pprint.pformat(sorted(self.fragments.items()))

    def __eq__(self, other):
        if isinstance(other, bytes):
            return self.tobytes() == other
        else:
            return self.tobytes() == other.tobytes()


class FragmentRegEx(object):
    def __init__(self, regexp, length):
        self.length = length if length else 1
        self.regexp = regexp

    def __len__(self):
        return self.length

class FragmentsOfRegexps(Fragments):
    def __init__(self, *args, **kargs):
        Fragments.__init__(self, *args, **kargs)
        self.regexp_by_position = {}

    def append(self, string, is_literal=True):
        assert isinstance(string, bytes)
        self.insert(self.current_offset, string, is_literal)

    def extend(self, iterable, is_literal=True):
        for string in iterable:
            assert isinstance(string, bytes)
            self.insert(self.current_offset, string, is_literal)

    def insert(self, position, string, is_literal=True):
        assert isinstance(string, bytes)
        if is_literal:
            regexp = re.escape(string)

        else:
            regexp = string
            string = b"x"

        Fragments.insert(self, position, string)

        self.regexp_by_position[position] = regexp

    def assemble_regexp(self):
        begin = 0
        result = []
        for p, regexp in sorted(self.regexp_by_position.items()):
            offset, string  = p, self.fragments[p]

            hole_length = (offset-begin)
            if hole_length > 0:
                result.append(("(?:.{%i})" % hole_length).encode('ascii'))

            result.append(regexp)
            begin = offset + len(string)

        return b''.join(result)




from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

from bisturi.six import integer_types

FROM_BEGIN = 0
FROM_END   = 2

def to_bytes(bytes_or_text):
    return (bytes_or_text if isinstance(bytes_or_text, bytes)
                          else bytes(bytes_or_text, 'ascii'))

class SeekableFile(bytes):
    def __new__(cls, file_open):
        self = bytes.__new__(cls)

        self._file = file_open
        self._file_length = self._calculate_file_length()

        return self

    def __getslice__(self, i, j):
        # deprecated, but still necessary to redefine the behavior of 'bytes'
        return self._slice(slice(i, j))

    def __getitem__(self, index):
        if isinstance(index, integer_types):
            self._seek(index)
            return self._file.read(1)

        elif isinstance(index, slice):
            return self._slice(index)

        else:
            raise TypeError("Invalid index/slice")

    def _slice(self, index):
        start, stop, step = index.indices(self._file_length)

        if step == 1:
            self._seek(start)
            return self._file.read(stop-start)

        else:
            return b"".join((self[i] for i in irange(start, stop, step)))

    def __str__(self):
        self._seek(0)
        return self._file.read()

    def _seek(self, offset):
        whence = FROM_BEGIN if offset >= 0 else FROM_END
        self._file.seek(offset, whence)

    def _calculate_file_length(self):
        self._file.seek(-1, FROM_END)
        length = self._file.tell() + 1

        self._file.seek(0, FROM_BEGIN)
        return length


def _string_as_seekable_file(s):
    from StringIO import StringIO
    return SeekableFile(file_open=StringIO(s))


import array
from bisturi.packet import Packet

import string
import binascii
import functools

_non_printable = set(range(256)) - set(ord(c) for c in string.printable)
_non_printable.union(ord(c) for c in string.whitespace if c != ' ')
_translate = ''.join('.' if i in _non_printable else chr(i) for i in range(256))
del _non_printable

def inspect(packet, indent="", current_level=0, aligned_to=8,
                truncate_values_to=8, max_items=4, max_depth=99999):
    inspection_too_deep = current_level > max_depth
    print("%s%s%s" % (indent, packet.__class__.__name__, " [truncated]"
                            if inspection_too_deep
                            else ""))

    if inspection_too_deep:
        return

    indent = "  " + indent

    inspect_recursive = functools.partial(inspect, current_level=current_level+1,
                                                   aligned_to=aligned_to,
                                                   truncate_values_to=truncate_values_to,
                                                   max_items=max_items,
                                                   max_depth=max_depth)
    ALIGNED_TO = aligned_to

    for name, _, _, _ in packet.get_fields():
        value = getattr(packet, name, None)
        indent_and_prefix = "%s%s = " % (indent, name)

        space = " " * (ALIGNED_TO - (len(indent_and_prefix) % ALIGNED_TO))

        if isinstance(value, array.array):
            value = value.tostring()

        if isinstance(value, basestring):
            l = len(value)
            MAX = truncate_values_to
            truncated = l > MAX
            truncated_value = value[:MAX]

            _value = []
            _value.append(' '.join(x.encode('hex') for x in truncated_value))
            _value.append(" " * ((MAX*3) - len(_value[-1])))
            _value.append("|%s|" % truncated_value.translate(_translate))
            if truncated:
                _value.append("[truncated]")

            _value.append("%i bytes" % l)

            value = "  ".join(_value)
            print("%s%s%s" % (indent_and_prefix, space, value))
        elif isinstance(value, Packet):
            print("%s%s:" % (indent, name))
            inspect_recursive(value, indent)

        elif isinstance(value, (list, tuple)):
            if value:
                truncated = False
                if isinstance(value[0], Packet):
                    truncated = len(value) > max_items
                else:
                    truncated = len(value) > max_items*8

                print("%s%s: %i items %s[" % \
                        (indent, name, len(value),
                        "[truncated] " if truncated else ""))
                if isinstance(value[0], Packet):
                    for subvalue in value[:max_items]:
                        inspect_recursive(subvalue, indent+"  ")

                else:
                    try:
                        print("%s%s" % \
                                (indent+"  ",
                                 " ".join(str(v) for v in value[:max_items*8])))
                    except:
                        print("%sunknow values" % (indent + "  "))

                print("%s]" % (indent, ))
            else:
                print("%s%s: []" % (indent, name))
        else:
            try:
                try:
                    print("%s%s%s" % (indent_and_prefix, space, valuea))
                except:
                    value = repr(value)
                    print("%s%s%s" % \
                            (indent_and_prefix, space, "unknow '%s'" % value))
            except:
                print("%s%s%s" % (indent_and_prefix, space, "unknow value"))




from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals
import bisturi.blocks
import copy, pprint

from bisturi.six import integer_types

__trace_enabled = False
__trace_indent = 0
def _trace(pargs=[], pattrs=[], presult=False):
    def decorator(method):
        def wrapper(self, *args, **kargs):
            global __trace_indent

            who = getattr(self, '__name__', self.__class__.__name__)
            indent = " " * __trace_indent
            print("{i}{who} {method}".format(i=indent, who=who,
                                             method=method.__name__))

            if pargs:
                print("{i}Args:".format(i=indent))
                for p in pargs:
                    try:
                        val = args[p]
                    except:
                        val = kargs[p]

                    print("{i}{arg}: {val}".format(i=indent, arg=p,
                                                    val=pprint.pformat(val)))

            __trace_indent += 1
            try:
                result = method(self, *args, **kargs)
            finally:
                __trace_indent -= 1

            if pattrs:
                print("{i}Attrs post-call:".format(i=indent))
                for p in pattrs:
                    val = getattr(self, p)
                    print("{i}{arg}: {val}".format(i=indent, arg=p,
                                                    val=pprint.pformat(val)))

            if presult:
                print("{i}Result: {val}".format(i=indent,
                                                    val=pprint.pformat(val)))

            return result

        if __trace_enabled:
            return wrapper
        else:
            return method
    return decorator

class PacketClassBuilder(object):
    def __init__(self, metacls, name, bases, attrs):
        self.metacls = metacls
        self.name = name
        self.bases = bases
        self.attrs = attrs

    def bisturi_configuration_default(self):
        return {}

    @_trace(pattrs=['bisturi_conf'])
    def make_configuration(self):
        defaults = self.bisturi_configuration_default()
        self.bisturi_conf = self.attrs.get('__bisturi__', defaults)

    def create_field_name_from_subpacket_name(self, subpacket_name):
        '''Helper method to transform names like CamelCase into camel_case'''
        name = subpacket_name[0].lower() + subpacket_name[1:]
        return "".join((c if c.islower() else "_"+c.lower()) for c in name)

    @_trace(pattrs=['attrs'])
    def create_fields_for_embebed_subclasses_and_replace_them(self):
        ''' Create Ref fields to refer to 'embebed' subclasses.
            Something like this:

            class A(Packet):
                class B(Packet):
                    pass

            is transformed into

            class A(Packet):
                b = Ref(B)

            See the method create_field_name_from_subpacket_name to know how
            the name 'B' was transformed into 'b'.
        '''

        from bisturi.packet import Packet
        from bisturi.field import Ref
        import inspect

        def is_a_packet_instance(name_and_field):
            _, field = name_and_field
            return inspect.isclass(field) and issubclass(field, Packet)

        names_and_subpackets = filter(is_a_packet_instance, self.attrs.items())
        subpackets_as_refs = [(self.create_field_name_from_subpacket_name(name),
                               Ref(prototype=subpacket, _is_a_subpacket_definition=True)) \
                                    for name, subpacket in names_and_subpackets]

        self.attrs.update(dict(subpackets_as_refs))

    @_trace(pattrs=['fields_in_class', 'original_fields_in_class'])
    def collect_the_fields_from_class_definition(self):
        ''' Collect the fields of the packet and sort them by creation time.
            Take something like this:

            class A(Packet):
                a = 1
                b = Int(2)
                c = Int(2)

            and collect [b->Int(2), c->Int(2)]
        '''
        from bisturi.field import Field

        def is_a_field_instance(name_and_field):
            _, field = name_and_field
            return isinstance(field, Field)

        def creation_time_of_field(name_and_field):
            _, field = name_and_field
            return field.ctime

        self.fields_in_class = filter(is_a_field_instance, self.attrs.items())
        self.fields_in_class = sorted(self.fields_in_class,
                                        key=creation_time_of_field)

        self.original_fields_in_class = list(self.fields_in_class)

    @_trace(pattrs=['fields'])
    def ask_to_each_field_to_describe_itself(self):
        ''' Ask to each field to describe itself. This should return for each
            field a list of names and fields which represent that original field.
            In most cases one field is described by only one field (itself) but
            there are cases where multiple field are needed.

            class A(Packet):
                a = Int(1)
                b = Int(2).at(0)

            The orignal list of fields should be [a->Int(1), b->Int(2)]
            but after the description of both fields we have a new list of fields:
                [a->Int(1), _shift_b_->Move(0), b->Int(2)]

            How each field is describe will depend of each field instance.
            See the method _describe_yourself of each Field subclass.
        '''
        self.fields = sum([field._describe_yourself(name, self.bisturi_conf) \
                            for name, field in self.fields_in_class], [])

    @_trace(pattrs=['slots'])
    def compile_fields_and_create_slots(self):
        ''' Compile each field, allowing them to optimize their pack/unpack
            methods.
            Also collect them and create the necessary slots to optimize the
            memory usage, then extend the slot list with the slots given by
            the user.
        '''

        def compile_field(position, name_and_field):
            _, field = name_and_field
            return field._compile(position, self.fields, self.bisturi_conf)

        additional_slots = self.bisturi_conf.get('additional_slots', [])
        self.slots = sum(map(compile_field, *zip(*enumerate(self.fields))), additional_slots)

    @_trace(pattrs=['slots'])
    def compile_descriptors_and_extend_slots(self):
        ''' Compile each field's descriptor if any and add their slots to the
            slot list.
        '''
        def has_descriptor(field):
            return field.descriptor is not None and hasattr(field.descriptor, '_compile')

        self.slots += sum((field.descriptor._compile(name, field.descriptor_name, self.bisturi_conf) for name, field in self.fields
               if has_descriptor(field)), [])


    @_trace()
    def lookup_pack_unpack_methods(self):
        ''' The list of fields is transformed in a list of tuples with the
            pack/unpack methods of each field ready to be called avoiding a
            further lookup.

            Take this [a->Int(1), b->Int(2)] into
                [(a, Int(1), a.pack, a.unpack),
                 (b, Int(2), b.pack, b.unpack),
                 ]
        '''
        self.fields = [(name, field, field.pack, field.unpack) for name, field in self.fields]

    @_trace(pattrs=['attrs'])
    def remove_fields_from_class_definition(self):
        ''' Remove from the class definition any field.
            Take this:
                class A(Packet):
                    a = 1
                    b = Int(2)

            and transform it into:
                class A(Packet):
                    a = 1
        '''
        for name, _ in self.original_fields_in_class:
            del self.attrs[name]

    @_trace(pattrs=['attrs', 'slots'])
    def add_descriptors_to_class_definition(self):
        ''' Add to the class definition any field's descriptor.
            It will replace a field by its descriptor.
            Take this:
                class A(Packet):
                    a = 1
                    b = Int(2).describe(Foo)

            and transform it into:
                class A(Packet):
                    a = 1
                    b = Foo()
        '''
        for name, field in self.fields:
            if field.descriptor:
                self.attrs[field.descriptor_name] = field.descriptor
                self.slots.remove(field.descriptor_name)

    @_trace()
    def collect_sync_methods_from_field_descriptors(self):
        self.sync_before_pack_methods = []
        self.sync_after_unpack_methods = []
        for name, field in self.fields:
            if field.descriptor:
                try:
                    self.sync_before_pack_methods.append(field.descriptor.sync_before_pack)
                except AttributeError:
                    pass

                try:
                    self.sync_after_unpack_methods.append(field.descriptor.sync_after_unpack)
                except AttributeError:
                    pass

    @_trace(pattrs=['metacls', 'name', 'bases', 'attrs', 'cls'])
    def create_class(self):
        ''' Create the class with the correct attributes and slots.
            If it is necessary, the original attributes (fields) can be access
            via the dictionary __bisturi__, key original_fields_in_class.
        '''
        self.bisturi_conf['original_fields_in_class'] = self.original_fields_in_class

        self.attrs['__slots__'] = self.slots
        self.attrs['__bisturi__'] = self.bisturi_conf

        self.cls = type.__new__(self.metacls, self.name, self.bases, self.attrs)

    @_trace()
    def add_get_fields_class_method(self):
        @classmethod
        def get_fields(cls):
            return self.fields

        self.cls.get_fields = get_fields

    @_trace()
    def add_sync_descriptor_class_methods(self):
        @classmethod
        def get_sync_before_pack_methods(cls):
            return self.sync_before_pack_methods

        @classmethod
        def get_sync_after_unpack_methods(cls):
            return self.sync_after_unpack_methods

        self.cls.get_sync_before_pack_methods = get_sync_before_pack_methods
        self.cls.get_sync_after_unpack_methods = get_sync_after_unpack_methods

    @_trace(pattrs=['am_in_debug_mode'])
    def check_if_we_are_in_debug_mode(self):
        ''' A class creation is in debug mode if one of its fields is
            a breakpoint (Bkpt).
        '''
        from bisturi.field import Bkpt
        self.am_in_debug_mode = any((isinstance(field, Bkpt) for _, field in self.fields))

    @_trace()
    def create_optimized_code(self):
        ''' Generate the optimized code for the pack and unpack methods and
            replace the original version for the optimized ones.

            The generation can be disabled is partially with the configuration
            flagos generate_for_pack/generate_for_unpack.
            And it is totally disabled if the class is in debug mode
            (see check_if_we_are_in_debug_mode)

            Optinally the generated code can be kept to manual inspection
            with the flag write_py_module in True.
        '''
        generate_by_default = True if not self.am_in_debug_mode else False

        generate_for_pack = self.cls.__bisturi__.get('generate_for_pack', generate_by_default)
        generate_for_unpack = self.cls.__bisturi__.get('generate_for_unpack', generate_by_default)

        write_py_module = self.cls.__bisturi__.get('write_py_module', False)

        bisturi.blocks.generate_code([(i, name_f[0], name_f[1]) for i, name_f in enumerate(self.fields)], self.cls, generate_for_pack, generate_for_unpack, write_py_module)

    @_trace()
    def get_packet_class(self):
        return self.cls

    @_trace()
    def create_collect_and_describe_the_field_list(self):
        ''' Given a class definition create any extra field necessary,
            then collect all of them and at last ask to each field to
            describe itself returning a final list of fields.
        '''
        self.create_fields_for_embebed_subclasses_and_replace_them()
        self.collect_the_fields_from_class_definition()
        self.ask_to_each_field_to_describe_itself()

    @_trace()
    def compile_fields_and_descriptors_and_create_slots(self):
        ''' Each field and each descriptor is compiled, optimized and
            added to the list of slots.
        '''
        self.compile_fields_and_create_slots()
        self.compile_descriptors_and_extend_slots()

    @_trace()
    def create_packet_class_and_add_its_special_methods(self):
        self.create_class()
        self.add_get_fields_class_method()
        self.add_sync_descriptor_class_methods()

    @_trace()
    def remove_fields_from_and_add_descriptors_to_class_definition(self):
        self.remove_fields_from_class_definition()
        self.add_descriptors_to_class_definition()

    @_trace()
    def optimize_methods(self):
        self.check_if_we_are_in_debug_mode()
        self.lookup_pack_unpack_methods()
        self.create_optimized_code()

class PacketSpecializationClassBuilder(PacketClassBuilder):
    def __init__(self, metacls, name, bases, attrs):
        from bisturi.packet import Packet

        self.super_class = attrs['__bisturi__']['specialization_of']
        assert isinstance(self.super_class, Packet)

        original_fields_in_superclass = self.super_class.__bisturi__['original_fields_in_class']
        specialized_fields = self.specialize_fields(attrs, original_fields_in_superclass)

        PacketClassBuilder.__init__(self, metacls, name, bases, specialized_attrs)

    def bisturi_configuration_default(self):
        return copy.deepcopy(self.super_class.__bisturi__)

    def specialize_fields(self, specialization_attrs, original_fields_in_superclass):
        specialized_fields = copy.deepcopy(original_fields_in_superclass)
        for attrname, attrvalue in specialization_attrs:
            if isinstance(attrvalue, Field) and attrname not in original_fields_in_superclass:
                raise Exception("You cannot add new fields like '%s'." % attrname)

            if isinstance(attrvalue, integer_types + (bytes, )) and attrname in original_fields_in_superclass:
                specialized_fields[attrname].default = attrvalue  # TODO the default or a constant??

            if isinstance(attrvalue, Field) and attrname in original_fields_in_superclass:
                attrvalue.ctime = original_fields_in_superclass[attrname].ctime # override the creation time to keep the same order
                specialized_fields[attrname] = attrvalue

        return specialized_fields

class MetaPacket(type):
    def __new__(metacls, name, bases, attrs):
        if name == 'Packet' and bases == (object,):
            attrs['__slots__'] = []
            return type.__new__(metacls, name, bases, attrs) # Packet base class

        specialization_of = attrs.get('__bisturi__', {}).get('specialization_of', None)
        if specialization_of:
            builder = PacketSpecializationClassBuilder(metacls, name, bases, attrs)

        else:
            builder = PacketClassBuilder(metacls, name, bases, attrs)

        builder.make_configuration()

        builder.create_collect_and_describe_the_field_list()
        builder.compile_fields_and_descriptors_and_create_slots()

        builder.collect_sync_methods_from_field_descriptors()

        builder.remove_fields_from_and_add_descriptors_to_class_definition()
        builder.create_packet_class_and_add_its_special_methods()

        builder.optimize_methods()

        cls = builder.get_packet_class()
        return cls




from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

from bisturi.field import Field, exec_once
from bisturi.deferred import UnaryExpr, BinaryExpr, NaryExpr, compile_expr_into_callable, defer_operations

from bisturi.six import integer_types

def normalize_raw_condition_into_a_callable(raw_condition):
    if callable(raw_condition):
        return raw_condition

    if isinstance(raw_condition, Field):
        raw_condition = convert_a_field_raw_condition_into_a_boolean_unary_expression(raw_condition)

    if isinstance(raw_condition, (UnaryExpr, BinaryExpr, NaryExpr)):
        raw_condition = compile_expr_into_callable(raw_condition)

    if callable(raw_condition):
        return raw_condition

    else:
        raise ValueError("The argument condition must be a callable, a field or an expression of fields but is '%s'" % (repr(raw_condition)))



def convert_a_field_raw_condition_into_a_boolean_unary_expression(a_field):
    assert isinstance(a_field, Field)

    # the order is important here, ask first for nonzero only then for len.
    # otherwise if 'a_field' is an Optional field and the Optional field's value
    # resolves to None, None doesn't have a __len__ method.
    # Of course, None doesn't have a __nonzero__ methods neither but
    # we use 'operator.truth' as the implementation of __nonzero__
    # and operator.truth(None) is well defined.
    truth_methods = ('__nonzero__', '__len__')

    for method in truth_methods:
        if hasattr(a_field, method):
            return getattr(a_field, method)()

    raise Exception("The field instance '%s' cannot be converted to a boolean value (it isn't an int nor a iterable)" % repr(a_field))


def normalize_count_condition_into_a_callable(count_raw_condition):
    if callable(count_raw_condition):
        return count_raw_condition

    if isinstance(count_raw_condition, integer_types):
        return lambda **k: count_raw_condition

    if isinstance(count_raw_condition, Field):
        field_name = count_raw_condition.field_name

        # aka int(count_raw_condition)
        count_raw_condition = lambda pkt, **k: getattr(pkt, field_name)

    if isinstance(count_raw_condition, (UnaryExpr, BinaryExpr, NaryExpr)):
        count_raw_condition = compile_expr_into_callable(count_raw_condition)

    if callable(count_raw_condition):
        return count_raw_condition

    else:
        raise ValueError("The 'count' condition must be a callable, a field or an expression of fields but is '%s'" % (repr(count_raw_condition)))


@defer_operations(allowed_categories=['sequence'])
class Sequence(Field):
    ''' Sequence of a fields (aka list of field).
        See the documentation of the method Field.repeated.
        '''
    def __init__(self, prototype, count=None, until=None, when=None, default=None, aligned=None):
        Field.__init__(self)
        assert isinstance(prototype, Field)

        if (count is None and until is None) or (count is not None and until is not None):
            raise ValueError("A sequence of fields (see the Field.repeated method) must have a count "
                             "of how many a field is repeated or a until condition to repeat "
                             "the field as long as the condition is false. "
                             "You must set one and only one of them.")

        self.ctime = prototype.ctime
        self.default = default if default is not None else []

        self.prototype_field = prototype
        self.aligned_to = aligned

        self.tmp = (count, until, when)

    @exec_once
    def _compile(self, position, fields, bisturi_conf):
        from bisturi.structural_fields import normalize_raw_condition_into_a_callable, \
                                      normalize_count_condition_into_a_callable

        slots = Field._compile_impl(self, position, fields, bisturi_conf)
        if self.aligned_to is None:
            self.aligned_to = bisturi_conf.get('align', 1)

        # XXX we are propagating the 'align' attribute (in bisturi_conf) to
        # the prototype packet. This is valid ...but inelegant
        # This happen in others fields like Optional and Ref
        self.seq_elem_field_name = "_seq_elem__"+self.field_name
        self.prototype_field.field_name = self.seq_elem_field_name
        self.prototype_field._compile(position=-1, fields=[], bisturi_conf=bisturi_conf)

        count, until, when = self.tmp

        self.when = None if when is None else normalize_raw_condition_into_a_callable(when)

        if count is None:
            self.get_how_many_elements  = None
            self.until_condition = normalize_raw_condition_into_a_callable(until)
        else:
            self.get_how_many_elements  = normalize_count_condition_into_a_callable(count)
            self.until_condition = None

        return slots + [self.seq_elem_field_name]

    def init(self, packet, defaults):
        Field.init(self, packet, defaults)
        self.prototype_field.init(packet, {})

    def unpack(self, pkt, raw, offset=0, **k):
        sequence = []
        setattr(pkt, self.field_name, sequence) # clean up the previous sequence (if any), 
                                                # so it can be used by the 'when' or 'until' callbacks

        count_elements = 1 if not self.get_how_many_elements else \
                           self.get_how_many_elements(pkt=pkt, raw=raw, offset=offset, **k)

        when = self.when
        if when and (count_elements <= 0 or not when(pkt=pkt, raw=raw, offset=offset, **k)):
            return offset

        seq_elem_field_name = self.seq_elem_field_name
        unpack = self.prototype_field.unpack
        append = sequence.append
        aligned_to = self.aligned_to
        for _ in range(count_elements):
            offset += (aligned_to - (offset % aligned_to)) % aligned_to
            offset =  unpack(pkt=pkt, raw=raw, offset=offset, **k)

            # because the unpack method must return a new instance or object each
            # time that it is called, we know that all the objects returned will be
            # different so we can append each one without worrying to be appending
            # several times the same object (and for that we don't need to create
            # a copy or anything else)
            append(getattr(pkt, seq_elem_field_name))

        until = self.until_condition
        should_continue = False if until is None else not until(pkt=pkt, raw=raw, offset=offset, **k)
        while should_continue:
            offset += (aligned_to - (offset % aligned_to)) % aligned_to
            offset =  unpack(pkt=pkt, raw=raw, offset=offset, **k)

            append(getattr(pkt, seq_elem_field_name))
            should_continue = not until(pkt=pkt, raw=raw, offset=offset, **k)

        return offset


    def pack(self, pkt, fragments, **k):
        sequence = getattr(pkt, self.field_name)
        seq_elem_field_name = self.seq_elem_field_name
        aligned_to = self.aligned_to
        pack = self.prototype_field.pack
        for val in sequence:
            setattr(pkt,  seq_elem_field_name, val)
            fragments.current_offset += (aligned_to - (fragments.current_offset % aligned_to)) % aligned_to
            pack(pkt, fragments, **k)

        return fragments

    def pack_regexp(self, pkt, fragments, **k):
        value = getattr(pkt, self.field_name)
        is_literal = not isinstance(value, Any)

        if is_literal:
            self.pack(pkt, fragments, **k)
        else:
            f = FragmentsOfRegexps()
            try:
                self.prototype_field.pack_regexp(pkt, f, **k)
                subregexp = f.assemble_regexp()
            except:
                subregexp = b".*"

            # TODO, fix this (fix the self.get_how_many_elements stuff)
            # (A)*
            fragments.append(b'(' + subregexp + b')*', is_literal=False)
            raise NotImplementedError("Not supported yet")
            if self.get_how_many_elements is None:
                fragments.append(b'(' + subregexp + b')*', is_literal=False)
            else:
                fragments.append(b'(' + subregexp + ("){%i}" % self.get_how_many_elements).encode('ascii'), is_literal=False)

        return fragments

    def repeated(self, *args, **kargs):
        r''' Nop, you cannot repeat a sequence (repeat twice):

             >>> from bisturi.packet import Packet
             >>> from bisturi.field  import Int, Data, Ref

             >>> class Buggy(Packet):                    # doctest: +ELLIPSIS
             ...    i = Int(1).repeated(2).repeated(4)
             Traceback (most recent call last):
             ...
             SyntaxError: You cannot repeat a sequence (more than one 'repeated' call is not allowed)...

             If you need this kind of 'list of list' behaviour you can do this:

             >>> class ListOfInts(Packet):
             ...    i = Int(1).repeated(2)

             >>> class NonBuggy(Packet):
             ...    i = Ref(ListOfInts).repeated(4)

             >>> raw = b'ABCDEFGH'
             >>> pkt = NonBuggy.unpack(raw)

             >>> [l_ints.i for l_ints in pkt.i]
             [[65, 66], [67, 68], [69, 70], [71, 72]]

             >>> pkt.pack() == raw
             True

             '''
        raise SyntaxError("You cannot repeat a sequence (more than one 'repeated' call is not allowed): something like Int(1).repeated(...).repeated(...). "
                          "See the documentation of Sequence.repeated for more info.")

    def when(self, *args, **kargs):
        r''' You cannot call 'when' of Sequence. Instead you can use the 'when'
             parameter of the Field.repeated method:

             >>> from bisturi.packet import Packet
             >>> from bisturi.field  import Int, Data, Ref

             >>> class Buggy(Packet):                    # doctest: +ELLIPSIS
             ...    t = Int(1)
             ...    i = Int(1).repeated(2).when(t)
             Traceback (most recent call last):
             ...
             SyntaxError: You cannot call 'when' of Sequence...

             Instead you should do something like:

             >>> class NonBuggy(Packet):
             ...    t = Int(1)
             ...    i = Int(1).repeated(2, when=t)

             '''
        raise SyntaxError("You cannot call 'when' of Sequence: something like Int(1).repeated(...).when(...). "
                          "Instead you need to use the 'when' parameter of 'repeated': Int(1).repeated(..., when=...).")


@defer_operations(allowed_categories='all') # we need all because the underlying object (value) can be an integer as well as a sequence 
class Optional(Field):
    ''' An optional field (aka the field or None).
        See the documentation of the method Field.when.
        '''
    def __init__(self, prototype, when, default=None):
        Field.__init__(self)
        assert isinstance(prototype, Field)

        self.ctime = prototype.ctime
        self.default = default

        self.prototype_field = prototype
        self.tmp = when

    @exec_once
    def _compile(self, position, fields, bisturi_conf):
        slots = Field._compile_impl(self, position, fields, bisturi_conf)
        self.opt_elem_field_name = "_opt_elem__"+self.field_name
        self.prototype_field.field_name = self.opt_elem_field_name
        self.prototype_field._compile(position=-1, fields=[], bisturi_conf=bisturi_conf)

        when = self.tmp
        del self.tmp

        self.when = normalize_raw_condition_into_a_callable(when)
        return slots + [self.opt_elem_field_name]

    def init(self, packet, defaults):
        Field.init(self, packet, defaults)
        self.prototype_field.init(packet, {})

    def unpack(self, pkt, raw, offset=0, **k):
        proceed = self.when(pkt=pkt, raw=raw, offset=offset, **k)

        opt_elem_field_name = self.opt_elem_field_name
        obj = None
        if proceed:
            offset = self.prototype_field.unpack(pkt=pkt, raw=raw, offset=offset, **k)

            obj = getattr(pkt, opt_elem_field_name)

        setattr(pkt, self.field_name, obj)
        return offset


    def pack(self, pkt, fragments, **k):
        obj = getattr(pkt, self.field_name)
        opt_elem_field_name = self.opt_elem_field_name
        if obj is not None:
            setattr(pkt,  opt_elem_field_name, obj)
            return self.prototype_field.pack(pkt, fragments, **k)

        else:
            return fragments

    def pack_regexp(self, pkt, fragments, **k):
        value = getattr(pkt, self.field_name)
        is_literal = not isinstance(value, Any)

        if is_literal:
            self.pack(pkt, fragments, **k)
        else:
            f = FragmentsOfRegexps()
            try:
                self.prototype_field.pack_regexp(pkt, f, **k)
                subregexp = f.assemble_regexp()
            except:
                subregexp = b".*"

            # (A)?
            fragments.append(b'(' + subregexp + b')?', is_literal=False)

        return fragments

    def repeated(self, *args, **kargs):
        r''' Nop, you cannot repeat an optional argument:

             >>> from bisturi.packet import Packet
             >>> from bisturi.field  import Int, Data, Ref

             >>> class Buggy(Packet):                    # doctest: +ELLIPSIS
             ...    t = Int(1)
             ...    i = Int(1).when(t == 0).repeated(4)
             Traceback (most recent call last):
             ...
             SyntaxError: You cannot repeat an optional argument...

             Instead you should do something like:

             >>> class NonBuggy(Packet):
             ...    t = Int(1)
             ...    i = Int(1).repeated(4, when = t == 0 )

             '''
        raise SyntaxError("You cannot repeat an optional argument: something like Int(1).when(...).repeated(...). "
                          "Instead you can pass the when condition to the 'repeated' method like Int(1).repeated(..., when=...). "
                          "See the arguments of Field.repeated for more info.")

    def when(self, *args, **kargs):
        r''' This is an optional field already, you cannot chain 'when' conditions:

             >>> from bisturi.packet import Packet
             >>> from bisturi.field  import Int, Data, Ref

             >>> class Buggy(Packet):                    # doctest: +ELLIPSIS
             ...    t = Int(1)
             ...    i = Int(1).when(t > 0).when(t < 10)
             Traceback (most recent call last):
             ...
             SyntaxError: You cannot make optional an already optional field...

             Instead you should do something like:

             >>> class NonBuggy(Packet):
             ...    t = Int(1)
             ...    i = Int(1).when((t > 0) & (t < 10))

             '''
        raise SyntaxError("You cannot make optional an already optional field: something like Int(1).when(...).when(...). "
                          "Instead you need to find a single condition that represent those two conditions into a single 'when' call.")

class Move(Field):
    def __init__(self, move_arg, movement_type):
        Field.__init__(self)
        self.move_arg = move_arg
        self.movement_type = movement_type
        self.default = b''

    def unpack(self, pkt, raw, offset=0, **k):
        if isinstance(self.move_arg, Field):
            move_value = getattr(pkt, self.move_arg.field_name)

        elif isinstance(self.move_arg, integer_types):
            move_value = self.move_arg

        else:
            assert callable(self.move_arg) # TODO the callable must have the same interface. currently recieve (pkt, raw, offset, **k) for unpack and (pkt, fragments, **k) for pack
            move_value = self.move_arg(pkt=pkt, raw=raw, offset=offset, **k)

        # TODO we need to disable this, the data may be readed by other field
        # in the future and then the packet will have duplicated data (but see pack())
        #setattr(pkt, self.field_name, raw[offset:move_value])
        if self.movement_type == 'absolute':
            return move_value
        elif self.movement_type == 'relative':
            return offset + move_value
        elif self.movement_type.startswith('align-'):
            if self.movement_type == 'align-global':
                start = 0
            elif self.movement_type == 'align-local':
                start = k['local_offset']
            else:
                raise Exception()

            return offset + ((move_value - ((offset-start) % move_value)) % move_value)
        else:
            raise Exception()

    def init(self, packet, defaults):
        pass

    def pack(self, pkt, fragments, **k):
        #garbage = getattr(pkt, self.field_name)

        if isinstance(self.move_arg, Field):
            move_value = getattr(pkt, self.move_arg.field_name)

        elif isinstance(self.move_arg, integer_types):
            move_value = self.move_arg

        else:
            assert callable(self.move_arg)
            move_value = self.move_arg(pkt=pkt, fragments=fragments, **k)

        # TODO because the "garbage" could be readed by another field in the future,
        # this may not be garbage and if  we try to put here, the other field will
        # try to put the same data in the same place and we get a collission.
        #fragments.append(garbage)
        if self.movement_type == 'absolute':
            fragments.current_offset = move_value
        elif self.movement_type == 'relative':
            fragments.current_offset += move_value
        elif self.movement_type.startswith('align-'):
            offset = fragments.current_offset
            if self.movement_type == 'align-global':
                start = 0
            elif self.movement_type == 'align-local':
                start = k['local_offset']
            else:
                raise Exception()

            fragments.current_offset += ((move_value - ((offset-start) % move_value)) % move_value)
        else:
            raise Exception()

        return fragments




from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

from bisturi.fragments import Fragments, FragmentsOfRegexps
from bisturi.pattern_matching import Any

try:
    import cPickle as pickle
except ImportError:
    import pickle

import copy, collections
import traceback, sys, re

import bisturi.packet_builder

from bisturi.six import with_metaclass

class PacketError(Exception):
    def __init__(self, was_error_found_in_unpacking_phase, field_name, packet_class_name, offset, original_error_message):
        Exception.__init__(self, "")
        self.original_traceback = "".join(traceback.format_exception(*sys.exc_info())[2:])

        self.was_error_found_in_unpacking_phase = was_error_found_in_unpacking_phase
        self.fields_stack = [(offset, field_name, packet_class_name)]
        self.original_error_message = original_error_message

    def add_parent_field_and_packet(self, offset, field_name, packet_class_name):
        self.fields_stack.append((offset, field_name, packet_class_name))

    def __str__(self):
        phase = "unpacking" if self.was_error_found_in_unpacking_phase else "packing"

        stack_details = []
        for offset, field_name, packet_class_name in reversed(self.fields_stack):
            offset_and_pkt_class = "    %08x %s" % (offset, packet_class_name)
            first_part_len = len(offset_and_pkt_class)

            space = " " * max(44 - first_part_len, 1)

            line = "%s%s.%s" % (offset_and_pkt_class, space, field_name)
            stack_details.append(line)

        stack_details = "\n".join(stack_details)

        closer_field_offset, closer_field_name, closer_packet_class_name = self.fields_stack[0]
        msg = "Error when %s the field '%s' of packet %s at %08x: %s\nPacket stack details: \n%s\nField's exception:\n%s" % (
                                 phase,
                                 closer_field_name, closer_packet_class_name,
                                 closer_field_offset,
                                 self.original_error_message,
                                 stack_details,
                                 self.original_traceback)

        return msg

class Packet(with_metaclass(bisturi.packet_builder.MetaPacket, object)):
    __bisturi__ = {}

    def __init__(self, _initialize_fields=True, **defaults):
        assert _initialize_fields in (True, False)
        if _initialize_fields:
            for field_name, field, _, _ in self.__class__.get_fields():
                field.init(self, defaults)
                try:
                    descriptor_name = field.descriptor_name
                    default_value = defaults[descriptor_name]
                    setattr(self, descriptor_name, default_value)
                except AttributeError:
                    pass
                except KeyError:
                    pass

    def as_prototype(self):
        return Prototype(self)

    @classmethod
    def unpack(cls, raw, offset=0, silent=False):
        if not isinstance(raw, bytes):
            raise ValueError("The raw parameter must be 'bytes', not '%s'." % type(raw))

        pkt = cls(_initialize_fields=False)
        try:
            pkt.unpack_impl(raw, offset, root=pkt)
            return pkt
        except PacketError as e:
            e.packet = pkt
            if silent:
                return None
            else:
                raise e
        except:
            if silent:
                return None
            else:
                raise


    def unpack_impl(self, raw, offset, **k):
        k['local_offset'] = offset
        try:
            for name, f, _, unpack in self.get_fields():
                offset = unpack(pkt=self, raw=raw, offset=offset, **k)
        except PacketError as e:
            e.add_parent_field_and_packet(offset, name, self.__class__.__name__)
            raise
        except Exception as e:
            raise PacketError(True, name, self.__class__.__name__, offset, str(e))

        [sync(self) for sync in self.get_sync_after_unpack_methods()]
        return offset

    def pack(self):
        fragments = Fragments()
        try:
            fragments = self.pack_impl(fragments, root=self)
            return fragments.tobytes()
        except PacketError as e:
            e.packet = self
            raise e


    def pack_impl(self, fragments, **k):
        [sync(self) for sync in self.get_sync_before_pack_methods()]
        k['local_offset'] = fragments.current_offset

        try:
            for name, f, pack, _ in self.get_fields():
                pack(pkt=self, fragments=fragments, **k)
        except PacketError as e:
            e.add_parent_field_and_packet(fragments.current_offset, name, self.__class__.__name__)
            raise
        except Exception as e:
            raise PacketError(False, name, self.__class__.__name__, fragments.current_offset, str(e))

        return fragments

    def as_regular_expression(self, debug=False):
        fragments = FragmentsOfRegexps()
        stack = []
        self.as_regular_expression_impl(fragments, stack)

        return re.compile(b"(?s)" + fragments.assemble_regexp(), re.DEBUG if debug else 0)


    def as_regular_expression_impl(self, fragments, stack):
        for name, f, pack, _ in self.get_fields():
            f.pack_regexp(self, fragments, stack=stack)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False

        for name, f, pack, _ in self.get_fields():
            if getattr(self, name) != getattr(other, name):
                return False

        return True


    def iterative_unpack(self, raw, offset=0, stack=None):
        raise NotImplementedError()
        for name, f, _, _ in self.get_fields():
            yield offset, name
            offset = f.unpack(pkt=self, raw=raw, offset=offset, stack=stack)

        yield offset, "."

class Prototype(object):
    def __init__(self, pkt):
        try:
            self.template = pickle.dumps(pkt, -1)
            pickle.loads(self.template) # sanity check
            self.clone = self._clone_from_pickle
        except Exception as e:
            self.template = copy.deepcopy(pkt)
            self.clone = self._clone_from_live_obj

    def clone(self):
        raise Exception()

    def _clone_from_pickle(self):
        return pickle.loads(self.template)

    def _clone_from_live_obj(self):
        return copy.deepcopy(self.template)




import sys

try:
    integer_types = (int, long)
except NameError:
    integer_types = (int, )

if sys.version_info[0] == 2:
    text_types = (unicode,)
else:
    text_types = (str,)


if sys.version_info[0] == 2:
    from_int_to_byte = chr
else:
    def from_int_to_byte(i):
        return bytes([i])


def with_metaclass(meta, *bases):
    """Create a base class with a metaclass."""
    # This requires a bit of explanation: the basic idea is to make a dummy
    # metaclass for one level of class instantiation that replaces itself with
    # the actual metaclass.
    class metaclass(meta):
        def __new__(cls, name, this_bases, d):
            return meta(name, bases, d)

    try:
        return type.__new__(metaclass, u'temporary_class', (), {})
    except TypeError:
        return type.__new__(metaclass, b'temporary_class', (), {})



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

try:
    from itertools import ifilter
except:
    ifilter = filter

from re import finditer, compile, escape
from functools import partial
from operator import eq as equals_to

class Any(object):
    def __init__(self, startswith=None, endswith=None, contains=None):
        if startswith == endswith == contains == None:  # most common case
            self.regexp = None
            self.__eq__ = self.eq_for_any
            self.__ne__ = self.ne_for_any

            return

        middle = b".*" if contains is None else (b".*%s.*" % escape(contains))

        self.regexp = b""
        if startswith is not None:
            self.regexp += escape(startswith)

        self.regexp += middle

        if endswith is not None:
            self.regexp += escape(endswith)

        self.regexp = compile(self.regexp)
        self.__eq__ = self.eq_for_regexp
        self.__ne__ = self.ne_for_regexp


    def __eq__(self, other):
        if self.regexp is None:
            return self.eq_for_any(other)
        else:
            return self.eq_for_regexp(other)

    def __ne__(self, other):
        if self.regexp is None:
            return self.ne_for_any(other)
        else:
            return self.ne_for_regexp(other)

    def eq_for_any(self, other):
        return True

    def ne_for_any(self, other):
        return False

    def eq_for_regexp(self, other):
        return bool(self.regexp.search(other))

    def ne_for_regexp(self, other):
        return not bool(self.regexp.search(other))

def anything_like(pkt_class):
    pkt = pkt_class()

    for field_name, field, _, _ in pkt_class.get_fields():
        setattr(pkt, field_name, Any())

    return pkt

def filter_like(pkt, iterable, scan_through_string_for_a_match=False):
    pattern = pkt.as_regular_expression()
    return ifilter(pattern.search if scan_through_string_for_a_match else pattern.match, iterable)

def filter(pkt, iterable, filter_with_regexp_first=True, filter_like_args={}):
    if filter_with_regexp_first:
        iterable = filter_like(pkt, iterable, **filter_like_args)

    cls = pkt.__class__
    equals_to_pkt = partial(equals_to, pkt)
    return ifilter(equals_to_pkt, (cls.unpack(r, silent=True) for r in iterable))



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

class Auto(object):
    def __init__(self, func):
        self.func = func

    def _compile(self, field_name, descriptor_name, bisturi_conf):
        self.iam_enabled_attr_name = "_is_descriptor_%s_enabled" % descriptor_name
        return [self.iam_enabled_attr_name]

    def __get__(self, instance, owner):
        if instance is None:
            return self

        iam_enabled = getattr(instance, self.iam_enabled_attr_name, True)
        if iam_enabled:
            return self.func(instance)
        else:
            real_value = getattr(instance, self.real_field_name)
            return real_value

    def __set__(self, instance, val):
        setattr(instance, self.iam_enabled_attr_name, False)
        setattr(instance, self.real_field_name, val)

    def __delete__(self, instance):
        setattr(instance, self.iam_enabled_attr_name, True)

    def sync_before_pack(self, instance):
        # this can be calculated or not, we don't care
        val = self.__get__(instance, type(instance))

        # we only care that the real field has the same value
        setattr(instance, self.real_field_name, val)


class AutoLength(Auto):
    def __init__(self, length_of):
        self.length_of = length_of
        Auto.__init__(self, self.calculate_length)

    def calculate_length(self, instance):
        return len(getattr(instance, self.length_of))




from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

import struct
import itertools
import hashlib
import os.path
import imp

def generate_code(fields, pkt_class, generate_for_pack, generate_for_unpack,
                                                            write_py_module):
    if not generate_for_pack and not generate_for_unpack:
        return

    grouped_by_variability = [(k, list(g)) for k, g in itertools.groupby(fields, lambda i_n_f: i_n_f[2].is_fixed)]
    codes = []
    for is_fixed, group in grouped_by_variability:
        if is_fixed:
            codes.extend(generate_code_for_fixed_fields(group))
        else:
            codes.append(generate_code_for_variable_fields(group))
    if generate_for_pack or generate_for_unpack:
        import_code = '''
from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

from struct import pack as StructPack, unpack as StructUnpack
from bisturi.fragments import Fragments
from bisturi.packet import PacketError

'''
    if generate_for_pack:
        pack_code = '''
def pack_impl(pkt, fragments, **k):
%(sync_descriptors_code)s
   k['local_offset'] = fragments.current_offset
   fields = pkt.get_fields()
   try:
%(blocks_of_code)s
   except PacketError as e:
      e.add_parent_field_and_packet(fragments.current_offset, name, pkt.__class__.__name__)
      raise e
   except Exception as e:
      raise PacketError(False, name, pkt.__class__.__name__, fragments.current_offset, str(e))

   return fragments
''' % {
      'blocks_of_code': indent("\n".join([c[0] for c in codes]), level=2),
      'sync_descriptors_code': generate_unrolled_code_for_descriptor_sync(
                                                pkt_class, sync_for_pack=True),
    }
    else:
        pack_code = ""

    if generate_for_unpack:
        unpack_code = ('''
from struct import pack as StructPack, unpack as StructUnpack
from bisturi.fragments import Fragments
from bisturi.packet import PacketError

def unpack_impl(pkt, raw, offset, **k):
   k['local_offset'] = offset
   fields = pkt.get_fields()
   try:
%(blocks_of_code)s
   except PacketError as e:
      e.add_parent_field_and_packet(offset, name, pkt.__class__.__name__)
      raise e
   except Exception as e:
      raise PacketError(True, name, pkt.__class__.__name__, offset, str(e))

%(sync_descriptors_code)s
   return offset
''' % {
      'blocks_of_code': indent("\n".join([c[1] for c in codes]), level=2),
      'sync_descriptors_code': generate_unrolled_code_for_descriptor_sync(pkt_class, sync_for_pack=False),
    })
    else:
        unpack_code = ""

    cookie_hash = hashlib.sha1()
    cookie_hash.update(pack_code.encode('utf-8'))
    cookie_hash.update(unpack_code.encode('utf-8'))
    cookie = cookie_hash.hexdigest()
    cookie_code = "BISTURI_PACKET_COOKIE = '%(cookie)s'\n" % {
      'cookie': cookie,
    }

    module_name = "_%s_pkt" % pkt_class.__name__
    module_filename = module_name + ".py"

    module = None
    if os.path.exists(module_filename):
        try:
            module = imp.load_source(module_name, module_filename)
        except ImportError:
            pass

    if module and hasattr(module, '__cached__'):
        module_compiled_filename = module.__cached__
    else:
        module_compiled_filename = module_name + ".pyc"

    if not module or getattr(module, 'BISTURI_PACKET_COOKIE', None) != cookie:
        if os.path.exists(module_compiled_filename):
            os.remove(module_compiled_filename)

        with open(module_filename, 'w') as module_file:
            module_file.write(import_code)
            module_file.write(cookie_code)
            module_file.write(pack_code)
            module_file.write(unpack_code)

        module = imp.load_source(module_name, module_filename)

        if module and hasattr(module, '__cached__'):
            module_compiled_filename = module.__cached__
        else:
            module_compiled_filename = module_name + ".pyc"

        if not write_py_module:
            try:
                os.remove(module_compiled_filename)
            except:
                pass
            os.remove(module_filename)

    from bisturi.packet import Packet
    if generate_for_pack and (pkt_class.pack_impl == Packet.pack_impl):
        pkt_class.pack_impl = module.pack_impl

    if generate_for_unpack and (pkt_class.unpack_impl == Packet.unpack_impl):
        pkt_class.unpack_impl = module.unpack_impl

def generate_unrolled_code_for_descriptor_sync(pkt_class, sync_for_pack):
    if sync_for_pack:
        sync_methods = pkt_class.get_sync_before_pack_methods()
        setup_code = "   sync_methods = pkt.get_sync_before_pack_methods()\n"
    else:
        sync_methods = pkt_class.get_sync_after_unpack_methods()
        setup_code = "   sync_methods = pkt.get_sync_after_unpack_methods()\n"

    if not sync_methods:
        return "   "

    sync_calls = '\n'.join('   sync_methods[%i](pkt)' % i \
                                        for i in range(len(sync_methods)))
    return setup_code + sync_calls


def generate_code_for_fixed_fields(fields):
    grouped_by_struct_code =  [(k, list(g)) for k, g in \
                                itertools.groupby(fields,
                                                  lambda i_n_f: i_n_f[2].struct_code is not None)]
    codes = []
    for has_struct_code, group in grouped_by_struct_code:
        if has_struct_code:
            grouped_by_endianness = [(k, list(g)) for k, g in \
                                        itertools.groupby(group,
                                                        lambda i_n_f: i_n_f[2].is_bigendian)]

            codes.extend([generate_code_for_fixed_fields_with_struct_code(g, k) \
                                for k, g in grouped_by_endianness])
        else:
            codes.append(generate_code_for_fixed_fields_without_struct_code(group))

    return codes

# TODO if is_bigendian  is  None means "don't care",
# no necessary means 'big endian (>)', so it should be joined
# with any other endianness
def generate_code_for_fixed_fields_with_struct_code(group, is_bigendian):
    fmt = ">" if is_bigendian else "<"
    fmt += "".join([f.struct_code for _, _, f in group])

    lookup_fields = " ".join([('pkt.%(name)s,' % {'name': name}) for _, name, _ in group])
    unpack_code = '''
name = "%(name)s"
next_offset = offset + %(advance)s
%(lookup_fields)s = StructUnpack("%(fmt)s", raw[offset:next_offset])
offset = next_offset
''' % {
         'lookup_fields': lookup_fields,
         'fmt': fmt,
         'advance': struct.calcsize(fmt),
         'name': ("between '%s' and '%s'" % (group[0][1], group[-1][1])) \
                    if len(group) > 1 else group[0][1],
      }

    pack_code = '''
name = "%(name)s"
fragments.append(StructPack("%(fmt)s", %(lookup_fields)s))
''' % {
         'lookup_fields': lookup_fields[:-1], # remove the last ","
         'fmt': fmt,
         'name': ("between '%s' and '%s'" % (group[0][1], group[-1][1])) \
                    if len(group) > 1 else group[0][1],
      }

    return pack_code, unpack_code

def generate_code_for_variable_fields(group):
    return (generate_code_for_loop_pack(group), generate_code_for_loop_unpack(group))

def generate_code_for_fixed_fields_without_struct_code(group):
    return (generate_code_for_loop_pack(group), generate_code_for_loop_unpack(group))

def generate_code_for_loop_pack(group):
    return ''.join(['''
name, _, pack, _ = fields[%(field_index)i]
pack(pkt=pkt, fragments=fragments, **k)
''' % {
      'field_index': field_index
   } for field_index in range(group[0][0], group[-1][0]+1)])

def generate_code_for_loop_unpack(group):
    return ''.join(['''
name, _, _, unpack = fields[%(field_index)i]
offset = unpack(pkt=pkt, raw=raw, offset=offset, **k)
''' % {
      'field_index': field_index
   } for field_index in range(group[0][0], group[-1][0]+1)])


def indent(code, level=1):
    i = "   " * level
    return "\n".join([i + line for line in code.split("\n")])




# https://packaging.python.org/en/latest/distributing.html
# https://github.com/pypa/sampleproject

from setuptools import setup
from codecs import open
from os import path, system

import sys

here = path.abspath(path.dirname(__file__))

system('''pandoc -o '%(dest_rst)s' '%(src_md)s' ''' % {
            'dest_rst': path.join(here, 'README.rst'),
            'src_md':   path.join(here, 'README.md'),
            })

with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

# load __version__, __doc__, _author, _license and _url
exec(open(path.join(here, 'bisturi', '__init__.py')).read())

system("find . -name '_[^_]*' -delete")
setup(
    name='bisturi',
    version=__version__,

    description=__doc__,
    long_description=long_description,

    url=_url,

    # Author details
    author=_author,
    author_email='use-github-issues@example.com',

    license=_license,

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',

        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',

        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',

        'Topic :: Software Development :: Code Generators',
        'Topic :: Software Development :: Compilers',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],

    python_requires='>=2.6',

    keywords='parse parsing binary network file compiler code generator',

    packages=['bisturi'],
)




from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

import time, struct, sys, copy, re

from bisturi.packet import Packet, Prototype
from bisturi.deferred import defer_operations, UnaryExpr, BinaryExpr, NaryExpr,\
                                    compile_expr_into_callable
from bisturi.pattern_matching import Any
from bisturi.fragments import FragmentsOfRegexps
from bisturi.util import to_bytes

from bisturi.six import integer_types, from_int_to_byte

def exec_once(m):
    ''' Force to execute the method m only once and save its result.
        The next calls will always return the same result,
        ignoring the parameters. '''
    def wrapper(self, *args, **kargs):
        try:
            return getattr(self, "_%s_cached_result" % m.__name__)
        except AttributeError as e:
            r = m(self, *args, **kargs)
            setattr(self, "_%s_cached_result" % m.__name__, r)
            return r

    return wrapper

class Field(object):
    ''' A field represent a single parsing unit. This is the superclass from
        where all the other field must inherit.

        A field will hold a raw configuration from it initialization (__init__)
        until the compilation of it (_compile method).
        Once you call _compile, the field will define the most optimum version
        for its pack and unpack methods.

        See the Packet class to get a more broad view about this.

        Any field instance will support the following public methods besides the
        pack/unpack:
            - repeated: to define a sequence of fields like Int(1).repeated(4)
              (see the Sequence field).
            - when: to define a field as optional like Int(1).when(foo == True)
              (see the Optional field).
            - at: to define where the field should start to pack/unpack
              (see the Move field).
            - aligned: a variant of at.

            '''
    def __init__(self):
        self.ctime = time.time()
        self.is_fixed = False
        self.struct_code = None
        self.is_bigendian = True

        self.move_arg = None
        self.movement_type = None

        self.descriptor = None
        self.descriptor_name = None

    def _describe_yourself(self, field_name, bisturi_conf):
        ''' Given a name and a configuration, set the name of this field
            and return the list of tuples that describe this field.
            Each tuple contains the name of the field and the value of the
            field.

            The most simple case is a description that consists in a list of one
            tuple
                [(my name, myself)]
            but it is possible that a field requires more than one tuple to
            describe.
            For example, Int(1).at(x) requires two tuples one about the movement
            (at) and the other about the int itself like:
                [(my move's name, Move(x)), (my name, myself)]
            '''
        from bisturi.structural_fields import Move

        self.field_name = field_name
        if self.move_arg is None and 'align' in bisturi_conf:
            self.aligned(to=bisturi_conf['align'])

        if self.descriptor:
            self.descriptor_name = field_name

            # The original field name will be used by the descriptor. We (field)
            # use a hidden field name instead
            self.field_name = "_described_%s" % field_name
            field_name = self.field_name

            # Notify to the descriptor both attribute names
            self.descriptor.descriptor_name = self.descriptor_name
            self.descriptor.real_field_name = self.field_name


        if self.move_arg is None:
            return [(field_name, self)]

        else:
            m = Move(self.move_arg, self.movement_type)
            m.field_name = "_shift_to_%s" % field_name
            return [(m.field_name, m), (field_name, self)]

    @exec_once
    def _compile(self, position, fields, bisturi_conf):
        ''' Realize all the optimizations available and return a list of names
            which will be the __slots__ of the packet class. This is the time
            to realize all the optimizations in terms of speed and memory as
            you can. '''
        # Dont call this from a subclass. Call _compile_impl directly.
        return self._compile_impl(position, fields, bisturi_conf)

    def _compile_impl(self, position, fields, bisturi_conf):
        slots = [self.field_name]
        if self.descriptor:
            slots.append(self.descriptor_name)

        return slots

    def init(self, packet, defaults):
        ''' Initialize the field based on the default.
            This must set a 'field_name' attribute in the packet.'''
        try:
            obj = defaults[self.field_name]
        except KeyError:
            obj = copy.deepcopy(self.default)

        setattr(packet, self.field_name, obj)

    def unpack(self, pkt, raw, offset, **k):
        raise NotImplementedError()

    def pack(self, pkt, fragments, **k):
        raise NotImplementedError()

    def unpack_noop(self, pkt, raw, offset, **k):
        ''' No-operation unpack function. Do nothing during the
            unpacking stage. '''
        return offset

    def pack_noop(self, pkt, fragments, **k):
        ''' No-operation pack function. Do nothing during the packing stage. '''
        return fragments

    def repeated(self, count=None, until=None, when=None, default=None,
                                                                aligned=None):
        r''' The sequence can be set to a fixed amount of elements with the
            'count' parameter which can be a number, a field, an expression of
            fields or even an arbitrary callable.
            In any case the parameter must be resolved to a positive integer.

            In the other hand, the sequence can set an 'until' condition.
            In this case the sequence will stop only when the until condition
            gives a true value.

            The 'count' and the 'until' parameter are exclusive: one and only
            one of them must be set.

            The 'when' condition can be used to make the whole sequence
            optional. If the when condition is not met, the sequence will be
            resolved to an empty list.

            >>> from bisturi.packet import Packet
            >>> from bisturi.field  import Int, Data, Ref

            >>> class Bag(Packet):
            ...     num = Int(1)
            ...     objects = Int(1).repeated(num)

            >>> class Box(Packet):
            ...     bags = Ref(Bag).repeated(until=lambda pkt, **k: pkt.bags[-1].num == 0)

            >>> pkt = Box(bags=[Bag(num=1, objects=[2]), Bag()])
            >>> len(pkt.bags)
            2
            >>> [(bag.num, bag.objects) for bag in pkt.bags]
            [(1, [2]), (0, [])]

            >>> pkt.pack() == b'\x01\x02\x00'
            True

            >>> raw = b'\x02\x01\x02\x01\x04\x00'
            >>> pkt = Box.unpack(raw)

            >>> len(pkt.bags)
            3
            >>> [(bag.num, bag.objects) for bag in pkt.bags]
            [(2, [1, 2]), (1, [4]), (0, [])]

            >>> pkt.pack() == raw
            True

            The 'aligned' parameter control how the elements of the sequence are
            packed (aligned) one each other.

            >>> class Room(Packet):
            ...     tight = Ref(Box).repeated(2, default=[Box(bags=[Bag()]), Box(bags=[Bag()])])
            ...     no_so_tight = Ref(Box).repeated(2, aligned=6, default=[Box(bags=[Bag()]), Box(bags=[Bag()])])

            >>> pkt = Room()
            >>> [sum((bag.objects for bag in box.bags), []) for box in pkt.tight]
            [[], []]
            >>> [sum((bag.objects for bag in box.bags), []) for box in pkt.no_so_tight]
            [[], []]

            >>> pkt.pack() == b'\x00\x00....\x00.....\x00'
            True

            >>> raw = b'\x01A\x00\x02BC\x00.....\x01A\x00...\x02BC\x00'
            >>> pkt = Room.unpack(raw)

            >>> [sum((bag.objects for bag in box.bags), []) for box in pkt.tight]
            [[65], [66, 67]]
            >>> [sum((bag.objects for bag in box.bags), []) for box in pkt.no_so_tight]
            [[65], [66, 67]]

            >>> pkt.pack() == raw
            True

            '''
        from bisturi.structural_fields import Sequence
        return Sequence(prototype=self, count=count, until=until, when=when,
                                            default=default, aligned=aligned)

    def when(self, condition, default=None):
        r''' A field can be set as optional based on a 'when' condition.
             This one can be a field, an expression of fields or an arbitrary
             callable that resolves to a boolean value: True if the field must
             be parsed or False otherwise.

             If a field is not parsed, None is used a the value for that field.

             The 'when' condition has no effect in a default packet neither
             during the packing phase.

             >>> from bisturi.packet import Packet
             >>> from bisturi.field  import Int, Data, Ref

             >>> class Example(Packet):
             ...     type = Int(1)
             ...     nonzero_msg = Data(2).when(type)
             ...     typeone_msg = Data(2).when(type == 1)

             >>> pkt = Example(nonzero_msg=b'X') # notice how all the fields are set...
             >>> (pkt.type, pkt.nonzero_msg, pkt.typeone_msg) # the when is ignored
             (0, 'X', None)

             >>> pkt.nonzero_msg = b'AB'
             >>> pkt.typeone_msg = b'CD'
             >>> pkt.pack() == b'\x00ABCD' # the when is ignored here too
             True

             >>> raw = b'\x00AB'
             >>> pkt = Example.unpack(raw) # here when is honored (both field...
             >>> (pkt.type, pkt.nonzero_msg, pkt.typeone_msg) # aren't parsed)
             (0, None, None)

             >>> pkt.pack() == b'\x00'
             True

             >>> raw = b'\x02AB'
             >>> pkt = Example.unpack(raw)
             >>> (pkt.type, pkt.nonzero_msg, pkt.typeone_msg)
             (2, 'AB', None)

             >>> raw = b'\x01ABCD'
             >>> pkt = Example.unpack(raw)
             >>> (pkt.type, pkt.nonzero_msg, pkt.typeone_msg)
             (1, 'AB', 'CD')

             >>> pkt.pack() == raw
             True

             '''

        from bisturi.structural_fields import Optional
        return Optional(prototype=self, when=condition, default=default)

    def at(self, position, movement_type='absolute'):
        self.move_arg = position
        self.movement_type = movement_type
        return self

    def aligned(self, to, local=False):
        self.move_arg = to
        self.movement_type = 'align-local' if local else 'align-global'
        return self

    def pack_regexp(self, pkt, fragments, **k):
        raise NotImplementedError("The pack_regexp method is not implemented for this field.")

    def describe(self, descriptor):
        self.descriptor = descriptor
        return self



@defer_operations(allowed_categories=['integer'])
class Int(Field):
    def __init__(self, byte_count=4, signed=False, endianness=None, default=0):
        Field.__init__(self)
        self.default = default
        self.byte_count = byte_count
        self.endianness = endianness
        self.is_signed = signed
        self.is_fixed = True

    @exec_once
    def _compile(self, position, fields, bisturi_conf):
        slots = Field._compile_impl(self, position, fields, bisturi_conf)

        if self.endianness is None:
            # try to get the default from the Packet class; big endian
            # by default
            self.endianness = bisturi_conf.get('endianness', 'big')

        self.is_bigendian = (self.endianness in ('big', 'network')) or \
                            (self.endianness == 'local' and sys.byteorder == 'big')

        if self.byte_count in (1, 2, 4, 8):
            code = {1:'B', 2:'H', 4:'I', 8:'Q'}[self.byte_count]
            if self.is_signed:
                code = code.lower()

            self.struct_code = code
            fmt = (">" if self.is_bigendian else "<") + code
            self.struct_obj = struct.Struct(fmt)

            self.pack, self.unpack = self._pack_fixed_and_primitive_size, \
                                        self._unpack_fixed_and_primitive_size

        else:
            self.struct_code = None
            self.base = 2**(self.byte_count*8)

            self.pack, self.unpack = self._pack_fixed_size, \
                                        self._unpack_fixed_size

        return slots

    def init(self, packet, defaults):
        setattr(packet, self.field_name, defaults.get(self.field_name,
                                                                self.default))
    def unpack(self, pkt, raw, offset=0, **k):
        raise NotImplementedError("This method should be implemented during the 'compilation' phase.")

    def pack(self, pkt, fragments, **k):
        raise NotImplementedError("This method should be implemented during the 'compilation' phase.")

    def _unpack_fixed_and_primitive_size(self, pkt, raw, offset=0, **k):
        next_offset = offset + self.byte_count
        integer = self.struct_obj.unpack(raw[offset:next_offset])[0]
        setattr(pkt, self.field_name, integer)

        return next_offset

    def _pack_fixed_and_primitive_size(self, pkt, fragments, **k):
        integer = getattr(pkt, self.field_name)
        raw = self.struct_obj.pack(integer)
        fragments.append(raw)

        return fragments

    def _unpack_fixed_size(self, pkt, raw, offset=0, **k):
        next_offset = offset + self.byte_count
        raw_data = raw[offset:next_offset]

        try:
            num = int.from_bytes(raw_data,
                        byteorder='big' if self.is_bigendian else 'little',
                        signed=self.is_signed)

        except AttributeError:
            if not self.is_bigendian:
                raw_data = raw_data[::-1]

            hexbytes = raw_data.encode('hex')
            num = int(hexbytes, 16)

            if self.is_signed and ord(raw_data[0]) > 127:
                num = -(self.base - num)

        setattr(pkt, self.field_name, num)
        return next_offset

    def _pack_fixed_size(self, pkt, fragments, **k):
        integer = getattr(pkt, self.field_name)

        try:
            data = integer.to_bytes(self.byte_count,
                            byteorder='big' if self.is_bigendian else 'little',
                            signed=self.is_signed)

        except AttributeError:
            num = (self.base + integer) if integer < 0 else integer

            xcode = "%%0%(count)ix" % {'count': self.byte_count*2}
            data = (xcode % num).decode('hex')

            if not self.is_bigendian:
                data = data[::-1]

        fragments.append(data)
        return fragments

    def pack_regexp(self, pkt, fragments, **k):
        value = getattr(pkt, self.field_name)
        is_literal = not isinstance(value, Any)

        if is_literal:
            self.pack(pkt, fragments, **k)
        else:
            fragments.append((".{%i}" % self.byte_count).encode('ascii'),
                                                            is_literal=False)

        return fragments

@defer_operations(allowed_categories=['sequence'])
class Data(Field):
    def __init__(self, byte_count=None, until_marker=None,
            include_delimiter=False, consume_delimiter=True, default=b''):
        Field.__init__(self)
        assert (byte_count is None and until_marker is not None) or \
                (until_marker is None and byte_count is not None)

        if until_marker is not None:
            if hasattr(until_marker, 'search'): # aka regex
                pattern = until_marker.pattern
                if not isinstance(pattern, bytes):
                    raise ValueError("The until marker is a regular expression which pattern is of type '%s' but it must be 'bytes'." % type(pattern))
            else:
                if not isinstance(until_marker, bytes):
                    raise ValueError("The until marker must be 'bytes' or a regular expression, not '%s'." % type(until_marker))

        if not isinstance(default, bytes):
            raise ValueError("The default must be 'bytes' not '%s'." %
                                                                type(default))

        self.default = default
        if not default and isinstance(byte_count, integer_types):
            self.default = b"\x00" * byte_count

        self.byte_count = byte_count
        self.until_marker = until_marker

        self.include_delimiter = include_delimiter
        self.delimiter_to_be_included = (self.until_marker if
                isinstance(self.until_marker, bytes) and not include_delimiter
                else b'')

        assert not (consume_delimiter == False and include_delimiter == True)
        self.consume_delimiter = consume_delimiter #XXX document this!
        self.is_fixed = isinstance(byte_count, integer_types)

    @exec_once
    def _compile(self, position, fields, bisturi_conf):
        slots = Field._compile_impl(self, position, fields, bisturi_conf)

        if self.byte_count is not None:
            if isinstance(self.byte_count, integer_types):
                self.struct_code = "%is" % self.byte_count
                self.unpack = self._unpack_fixed_size

            elif isinstance(self.byte_count, Field):
                self.unpack = self._unpack_variable_size_field

            elif callable(self.byte_count):
                self.unpack = self._unpack_variable_size_callable

            elif isinstance(self.byte_count, (UnaryExpr, BinaryExpr, NaryExpr)):
                self.byte_count = compile_expr_into_callable(self.byte_count)
                self.unpack = self._unpack_variable_size_callable

            else:
                assert False

        else:
            self._search_buffer_length = bisturi_conf.get('search_buffer_length')
            if self._search_buffer_length is not None:
                # the length can be 0 or None (means infinite) or a positive number
                assert self._search_buffer_length >= 0

            if isinstance(self.until_marker, bytes):
                self.unpack = self._unpack_with_string_marker

            elif hasattr(self.until_marker, 'search'):
                self.unpack = self._unpack_with_regexp_marker

            else:
                assert False

        return slots

    def init(self, packet, defaults):
        setattr(packet, self.field_name, defaults.get(self.field_name,
                                                                self.default))

    def unpack(self, pkt, raw, offset=0, **k):
        raise NotImplementedError("This method should be implemented during the 'compilation' phase.")

    def pack(self, pkt, fragments, **k):
        r = getattr(pkt, self.field_name) + self.delimiter_to_be_included
        fragments.append(r)
        return fragments

    def _unpack_fixed_size(self, pkt, raw, offset=0, **k):
        byte_count = self.byte_count
        next_offset = offset + byte_count

        chunk = raw[offset:next_offset]
        if len(chunk) != byte_count:
            raise Exception("Unpacked %i bytes but expected %i" %
                                                    (len(chunk), byte_count))

        setattr(pkt, self.field_name, chunk)
        return next_offset

    def _unpack_variable_size_field(self, pkt, raw, offset=0, **k):
        byte_count = getattr(pkt, self.byte_count.field_name)
        next_offset = offset + byte_count

        chunk = raw[offset:next_offset]
        if len(chunk) != byte_count:
            raise Exception("Unpacked %i bytes but expected %i" %
                                                    (len(chunk), byte_count))

        setattr(pkt, self.field_name, chunk)
        return next_offset

    def _unpack_variable_size_callable(self, pkt, raw, offset=0, **k):
        byte_count = self.byte_count(pkt=pkt, raw=raw, offset=offset, **k)
        next_offset = offset + byte_count

        chunk = raw[offset:next_offset]
        if len(chunk) != byte_count:
            raise Exception("Unpacked %i bytes but expected %i" %
                                                    (len(chunk), byte_count))

        setattr(pkt, self.field_name, chunk)
        return next_offset

    def _unpack_with_string_marker(self, pkt, raw, offset=0, **k):
        until_marker = self.until_marker

        if self._search_buffer_length:
            max_next_offset_allowed = offset + self._search_buffer_length
            search_buffer = raw[offset:max_next_offset_allowed]
        else:
            search_buffer = raw[offset:]

        count = search_buffer.find(until_marker)
        assert count >= 0

        extra_count = 0
        if self.include_delimiter:
            count += len(until_marker)
        else:
            self.delimiter_to_be_included = until_marker
            if self.consume_delimiter:
                extra_count = len(until_marker)

        next_offset = offset + count
        setattr(pkt, self.field_name, raw[offset:next_offset])

        return next_offset + extra_count

    def _unpack_with_regexp_marker(self, pkt, raw, offset=0, **k):
        until_marker = self.until_marker

        if self._search_buffer_length:
            max_next_offset_allowed = offset + self._search_buffer_length
            search_buffer = raw[offset:max_next_offset_allowed]
        else:
            search_buffer = raw[offset:]

        extra_count = 0
        if until_marker.pattern == b"$":    # shortcut
            count = len(raw) - offset
        else:
            match = until_marker.search(search_buffer, 0) #XXX should be (raw, offset) or (raw[offset:], 0) ? See the method search in the module re
            if match:
                if self.include_delimiter:
                    count = match.end()
                else:
                    count = match.start()
                    if self.consume_delimiter:
                        extra_count = match.end()-count
                    self.delimiter_to_be_included = match.group()
            else:
                assert False

        next_offset = offset + count
        setattr(pkt, self.field_name, raw[offset:next_offset])

        return next_offset + extra_count

    def pack_regexp(self, pkt, fragments, **k):
        value = getattr(pkt, self.field_name)
        is_literal = not isinstance(value, Any)

        if is_literal:
            self.pack(pkt, fragments, **k)

        else:
            custom_regexp = (value.regexp.pattern if value.regexp is not None
                                                  else b".*")
            if self.byte_count is not None:
                if isinstance(self.byte_count, integer_types):
                    byte_count = self.byte_count

                elif isinstance(self.byte_count, Field):
                    byte_count = getattr(pkt, self.byte_count.field_name)

                elif callable(self.byte_count):
                    try:
                        byte_count = self.byte_count(pkt=pkt, **k)
                    except Exception as e:
                        byte_count = None

                if byte_count is not None:
                    # TODO ignoring the custom regexp!!
                    fragments.append((".{%i}" % byte_count).encode('ascii'),
                                                            is_literal=False)
                else:
                    fragments.append(custom_regexp, is_literal=False)

            else:
                endswith = (re.escape(self.until_marker) if
                                        isinstance(self.until_marker, bytes)
                                        else self.until_marker.pattern)
                fragments.append(custom_regexp + endswith, is_literal=False)

        return fragments

class Ref(Field):
    r'''Reference to another packet description. This field allows to composite
        different packet descriptions.

        The prototype parameter must be a packet class, a packet instance or a
        callable.
        If it is the latter, the callable must return a packet instance or a
        field instance each time that it is called.
        Because of that, the callable must to return a new packet/field instance
        each time, you cannot return the same object twice.

        The callable will be called during the unpack stage but it can also be
        called during the pack stage to determinate how to pack a value when
        this one is not a packet instance (let's say that you are referencing to
        an Int field and I have the value 42, I need to call the callable to
        get an Int instance to pack the 42).
        In this case the callable will be call with the parameter packing=True.

        If the prototype is a packet, then the default it is not needed becasue
        we can use the prototype as a default value for the field.
        But if it is a callable, the default is mandatory otherwise we cannot
        know how to create a default value for the field.

        >>> from bisturi.packet import Packet
        >>> from bisturi.field  import Int, Data, Ref

        >>> class Point(Packet):
        ...     x = Int(1)
        ...     y = Int(1)

        >>> class Line(Packet):
        ...     begin = Ref(Point(x=1, y=2)) # prototype is used as the default
        ...     end   = Ref(Point) # shortcut for Ref(Point())
        ...
        ...     extra = Ref(lambda **k: Point(), default=Point(y=7)) # default is mandatory

        >>> pkt = Line()
        >>> (pkt.begin.x, pkt.begin.y)
        (1, 2)
        >>> (pkt.end.x, pkt.end.y)
        (0, 0)
        >>> (pkt.extra.x, pkt.extra.y)
        (0, 7)

        >>> pkt.pack() == b'\x01\x02\x00\x00\x00\x07'
        True

        >>> raw = b'\x01\x02\x03\x04\x05\x06'
        >>> pkt = Line.unpack(raw)

        >>> (pkt.begin.x, pkt.begin.y)
        (1, 2)
        >>> (pkt.end.x, pkt.end.y)
        (3, 4)
        >>> (pkt.extra.x, pkt.extra.y)
        (5, 6)

        >>> pkt.pack() == raw
        True


        If the embeb parameter is True, the prototype must be a packet.
        With this flag all the fields of the prototype are borrow to the
        current packet.
        The embeb feature is quite experimental and has a few quirks:

        >>> class Point3D(Packet):
        ...     point_2d = Ref(Point(x=1, y=2), embeb=True)
        ...     z = Int(1)

        >>> pkt = Point3D(x=7)
        >>> (pkt.x, pkt.y, pkt.z) # XXX this should be (7, 2, 0) but...
        (7, 0, 0)

        >>> pkt.x = 8
        >>> pkt.point_2d.y = 9
        >>> pkt.pack() == b'\x08\x00\x00' # XXX but it should be 08 09 00
        True

        >>> raw = b'\x01\x02\x03'
        >>> pkt = Point3D.unpack(raw)
        >>> (pkt.x, pkt.y, pkt.z)
        (1, 2, 3)

        >>> pkt.pack() == b'\x01\x02\x03'
        True

        '''

    def __init__(self, prototype, default=None, embeb=False,
                                            _is_a_subpacket_definition=False):
        Field.__init__(self)

        if _is_a_subpacket_definition:
            self.ctime = prototype.get_fields()[0][1].ctime

        self.default = default

        if isinstance(prototype, type):
            # get an object, this allows write Ref(PacketClass)
            # instead of Ref(PacketClass())
            prototype = prototype()

        if not isinstance(prototype, Packet) and not callable(prototype) and \
                not isinstance(prototype, (UnaryExpr, BinaryExpr, NaryExpr)):
            raise ValueError("The prototype of a Ref field must be a packet (class or instance), an expression of fields or a callable that should return a Field or a Packet.")

        self._lets_find_a_nice_default(prototype, default)

        if embeb and not isinstance(prototype, Packet):
            raise ValueError("The prototype must be a Packet if you want to embeb it.")

        self.prototype = prototype
        self.embeb = embeb

    def _lets_find_a_nice_default(self, prototype, default):
        if callable(prototype) or isinstance(prototype, (UnaryExpr, BinaryExpr,
                                                         NaryExpr)):
            if default is None:
                raise ValueError("If your are using an expression of fields or a callable as the prototype of Ref I need a default object.")

            self.default = default

        elif isinstance(prototype, Packet):
            if default is not None:
                raise ValueError("We don't need a default object, we will be using the prototype object instead.")

            self.default = copy.deepcopy(prototype)

        else:
            assert False

    def _describe_yourself(self, field_name, bisturi_conf):
        desc = Field._describe_yourself(self, field_name, bisturi_conf)
        if self.embeb:
            desc.extend([(fname, f) for fname, f, _, _ in self.prototype.get_fields()])

        return desc

    @exec_once
    def _compile(self, position, fields, bisturi_conf):
        slots = Field._compile_impl(self, position, fields, bisturi_conf)

        self.position = position

        prototype = self.prototype
        if isinstance(prototype, Packet):
            self.unpack = self._unpack_referencing_a_packet
            self.pack   = self._pack_referencing_a_packet
            self.prototype   = prototype.as_prototype()
            self.proto_class = prototype.__class__

        else:
            assert callable(prototype) or isinstance(prototype, (UnaryExpr,
                                                                 BinaryExpr,
                                                                 NaryExpr))
            from bisturi.structural_fields import normalize_raw_condition_into_a_callable
            self.prototype = normalize_raw_condition_into_a_callable(prototype)
            prototype = self.prototype

            if isinstance(self.default, Prototype):
                self.default = self.default.as_prototype()

            self.unpack = self._unpack_using_callable
            self.pack   = self._pack_with_callable

        if self.embeb:
            assert isinstance(prototype, Packet)
            self.pack   = self.pack_noop
            self.unpack = self.unpack_noop

        assert not isinstance(self.prototype, Packet)
        assert isinstance(self.prototype, Prototype) or callable(self.prototype)
        return slots

    def init(self, packet, defaults):
        # we use our prototype to get a valid default if the prototype is not a
        # callable in the other case, we use self.default.
        prototype = self.prototype
        if isinstance(prototype, Prototype):
            if self.field_name not in defaults:
                defaults[self.field_name] = prototype.clone()

            Field.init(self, packet, defaults)

        else:
            assert callable(self.prototype)
            default = self.default
            if isinstance(default, Prototype):
                if self.field_name not in defaults:
                    defaults[self.field_name] = default.clone()

            Field.init(self, packet, defaults)

    def _unpack_using_callable(self, pkt, raw, offset=0, **k):
        referenced = self.prototype(pkt=pkt, raw=raw, offset=offset, **k)

        if isinstance(referenced, Field):
            referenced.field_name=self.field_name
            referenced._compile(position=self.position, fields=[], bisturi_conf={})
            referenced.init(pkt, {})

            return referenced.unpack(pkt=pkt, raw=raw, offset=offset, **k)

        assert isinstance(referenced, Packet)

        setattr(pkt, self.field_name, referenced)
        return referenced.unpack_impl(raw, offset, **k)

    def _pack_with_callable(self, pkt, fragments, **k):
        # this can be a Packet or can be anything (but not a Field: it could be
        # a 'int' for example but not a 'Int')
        obj = getattr(pkt, self.field_name)
        if isinstance(obj, Packet):
            return obj.pack_impl(fragments=fragments, **k)

        # we try to know how to pack this value
        assert callable(self.prototype)
        # TODO add more parameters, like raw=partial_raw
        referenced = self.prototype(pkt=pkt, fragments=fragments, packing=True, **k)

        if isinstance(referenced, Field):
            referenced.field_name = self.field_name
            referenced._compile(position=self.position, fields=[], bisturi_conf={})
            #referenced.init(pkt, {})

            return referenced.pack(pkt, fragments, **k)

        # well, we are in a dead end: the 'obj' object IS NOT a Packet,
        # it is a "primitive" value however, the 'referenced' object IS
        # a Packet and we cannot do anything else
        raise NotImplementedError("I have a value to pack of type '%s' and because it is not a Packet instance I don't know how to pack it. The prototype of this Ref field is a callable so I called it hoping to receive a Field instance to show me how to pack the value but instead I received a '%s' so I'm stuck." % (type(obj), type(referenced)))


    def _unpack_referencing_a_packet(self, pkt, **k):
        p = self.proto_class(_initialize_fields=False)
        setattr(pkt, self.field_name, p)
        return p.unpack_impl(**k)

    def _pack_referencing_a_packet(self, pkt, fragments, **k):
        return getattr(pkt, self.field_name).pack_impl(fragments=fragments, **k)


@defer_operations(allowed_categories=['integer'])
class Bits(Field):
    class ByteBoundaryError(Exception):
        def __init__(self, str):
            Exception.__init__(self, str)

    def __init__(self, bit_count, default=0):
        Field.__init__(self)
        self.default = default
        self.mask = (2**bit_count)-1
        self.bit_count = bit_count

        self.iam_first = self.iam_last = False

    @exec_once
    def _compile(self, position, fields, bisturi_conf):
        slots = Field._compile_impl(self, position, fields, bisturi_conf)

        if position == 0 or not isinstance(fields[position-1][1], Bits):
            self.iam_first = True

        if position == len(fields)-1 or not isinstance(fields[position+1][1], Bits):
            self.iam_last = True

        if self.iam_last:
            cumshift = 0
            I = Int()
            self.members = []
            for n, f in reversed(fields[:position+1]):
                if not isinstance(f, Bits):
                    break

                f.shift = cumshift
                f.mask = ((2**f.bit_count) - 1) << f.shift
                f.I = I

                self.members.append((n, f.bit_count))

                cumshift += f.bit_count
                del f.bit_count

            self.members.reverse()
            name_of_members, bit_sequence = zip(*self.members)

            if not (cumshift % 8 == 0):
                raise Bits.ByteBoundaryError("Wrong sequence of bits: %s with total sum of %i (not a multiple of 8)." % (str(list(bit_sequence)), cumshift))

            I.byte_count = cumshift // 8
            fname = "_bits__"+"_".join(name_of_members)
            I.field_name = fname
            I._compile(position=-1, fields=[], bisturi_conf={})

            slots.append(fname)
        return slots


    def init(self, packet, defaults):
        if self.iam_first:
            setattr(packet, self.I.field_name, 0)

        setattr(packet, self.field_name, defaults.get(self.field_name, self.default))

    def unpack(self, pkt, raw, offset=0, **k):
        if self.iam_first:
            offset = self.I.unpack(pkt, raw, offset, **k)

        I = getattr(pkt, self.I.field_name)

        setattr(pkt, self.field_name, (I & self.mask) >> self.shift)
        return offset

    def pack(self, pkt, fragments, **k):
        I = getattr(pkt, self.I.field_name)
        setattr(pkt, self.I.field_name, ((getattr(pkt, self.field_name) << self.shift) & self.mask) | (I & (~self.mask)))

        if self.iam_last:
            return self.I.pack(pkt, fragments=fragments, **k)
        else:
            return fragments

    def pack_regexp(self, pkt, fragments, **k):
        if self.iam_last:
            bits = []
            for name, bit_count in self.members:
                is_literal = not isinstance(getattr(pkt, name), Any)

                if is_literal:
                    b = bin(getattr(pkt, name))[2:]
                    zeros = bit_count-len(b)

                    bits.append("0" * zeros)
                    bits.append(b)
                else:
                    bits.append("x" * bit_count)

            bits  = "".join(bits)
            bytes_ = [bits[0+(i*8):8+(i*8)] for i in range(len(bits)//8)]

            for byte in bytes_:
                if byte == "x" * 8:
                    # xxxx xxxx pattern (all dont care)
                    fragments.append(b".{1}", is_literal=False)
                else:
                    first_dont_care = byte.find("x")
                    if first_dont_care == -1:
                        # 0000 0000 pattern (all fixed)
                        char = from_int_to_byte(int(byte, 2))
                        fragments.append(char, is_literal=True)

                    elif byte[first_dont_care:] == "x" * len(byte[first_dont_care:]):
                        # 00xx xxxx pattern (lower dont care)
                        dont_care_bits = len(byte[first_dont_care:])

                        lower_bin = byte[:first_dont_care] + ("0"*dont_care_bits)
                        higher_bin = byte[:first_dont_care] + ("1"*dont_care_bits)
                        lower_char = from_int_to_byte(int(lower_bin,  2))
                        higher_char = from_int_to_byte(int(higher_bin, 2))

                        lower_literal = re.escape(lower_char)
                        higher_literal = re.escape(higher_char)

                        # [lower-higher]
                        fragments.append(b'[' + \
                                          lower_literal + \
                                          b'-' + \
                                          higher_literal + \
                                          b']', is_literal=False)

                    else:
                        # 00xx x0x0 pattern (mixed pattern)
                        all_patterns   = range(256)
                        fixed_pattern  = int(byte.replace("x", "0"), 2)
                        dont_care_mask = int(byte.replace("1", "0").replace("x", "1"), 2)

                        mixed_patterns = sorted(
                                            set((p & dont_care_mask) | \
                                            fixed_pattern for p in all_patterns))

                        # [ABCD....]
                        literal_patterns = (re.escape(from_int_to_byte(p)) \
                                                        for p in mixed_patterns)
                        fragments.append(b'[' + \
                                         b''.join(literal_patterns) + \
                                         b']', is_literal=False)

        return fragments


class Bkpt(Field):
    def __init__(self):
        Field.__init__(self)

    def init(self, packet, defaults):
        pass

    def unpack(self, pkt, raw, offset=0, **k):
        import pdb
        pdb.set_trace()
        return offset

    def pack(self, pkt, fragments, **k):
        import pdb
        pdb.set_trace()
        return fragments

    def pack_regexp(self, pkt, fragments, **k):
        return fragments

class Em(Field):
    def __init__(self):
        Field.__init__(self)

    def init(self, packet, defaults):
        pass

    @exec_once
    def _compile(self, position, fields, bisturi_conf):
        return []

    def unpack(self, pkt, raw, offset=0, **k):
        return offset

    def pack(self, pkt, fragments, **k):
        fragments.append(b"")
        return fragments




from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

from bisturi.six import text_types

import collections, functools, operator

def if_true_then_else(condition, possible_values):
    value_if_true, value_if_false = possible_values
    return value_if_true if bool(condition) else value_if_false

def choose(index, options):
    return options[index]

AllCategories = ['integer', 'sequence']
BinaryOperationsByCategory = {
        'integer': [
            # arith ------------------------------------
            operator.add,         operator.sub,
            operator.mul,
            operator.truediv,     operator.floordiv,
            operator.mod,         operator.pow,

            # cmp --------------------------------------
            operator.le,          operator.lt,
            operator.ge,          operator.gt,

            # eq ---------------------------------------
            operator.eq,          operator.ne,

            # logical ----------------------------------
            operator.and_,        operator.or_,
            operator.xor,
            operator.rshift,      operator.lshift,
            ],

        'sequence': [
            # eq ---------------------------------------
            operator.eq,          operator.ne,

            # indexing ---------------------------------
            operator.getitem,
            ],
        }

BinaryReverseOperationsByCategory = {
        'integer': [
            # arith ------------------------------------
            operator.add,         operator.sub,
            operator.mul,
            operator.truediv,     operator.floordiv,
            operator.mod,         operator.pow,

            # logical ----------------------------------
            operator.and_,        operator.or_,
            operator.xor,
            operator.rshift,      operator.lshift,
            ],

        'sequence': [
            ],
        }

UnaryOperationsByCategory = {
        'integer': [
            # arith ------------------------------------
            operator.neg,

            # logical ----------------------------------
            operator.inv,

            # value ------------------------------------
            operator.truth,
            ],

        'sequence': [
            # length -----------------------------------
            len,
            ],
        }


NaryExpr = collections.namedtuple('NaryExpr', ['l', 's', 'm', 'op'])
BinaryExpr = collections.namedtuple('BinaryExpr', ['l', 'r', 'op'])
UnaryExpr = collections.namedtuple('UnaryExpr', ['r', 'op'])

class Operations(object):
    def __init__(self):
        self.ops = []

    def append(self, num_arguments, operation, level, operation_name=None):
        if operation_name is None:
            operation_name = repr(operation)

        self.ops.append((num_arguments, operation, level, operation_name))

    def as_list(self):
        return [(num_arguments, operation) for num_arguments, operation, _, _ in self.ops]


def compile_expr(root_expr, ops=None, level=0, verbose=False):
    from bisturi.field import Field
    next_level = level + 1

    if ops is None:
        ops = Operations()

    if not isinstance(root_expr, (UnaryExpr, BinaryExpr, NaryExpr, Field)):
        ops.append(0, lambda pkt, *vargs, **kargs: root_expr, level, 'literal-value ' + repr(root_expr))

    elif isinstance(root_expr, NaryExpr):
        r, l, m, op = root_expr

        compile_expr(r, ops, level=next_level)

        assert l or m
        assert not (l and m)

        if l:
            n = len(l)
            for value in l:
                compile_expr(value, ops, level=next_level)

            ops.append(n, lambda *vargs: vargs, level, 'arg-list')

        else:
            n = len(m)
            keys, values = zip(*m.items())
            for value in values:
                compile_expr(value, ops, level=next_level)

            ops.append(n, lambda *vargs: dict(zip(keys, vargs)), level, 'arg-mapping')

        ops.append(2, op, level)

    elif isinstance(root_expr, BinaryExpr):
        l, r, op = root_expr
        compile_expr(l, ops, level=next_level)
        compile_expr(r, ops, level=next_level)
        ops.append(2, op, level)

    elif isinstance(root_expr, UnaryExpr):
        r, op = root_expr
        compile_expr(r, ops, level=next_level)
        ops.append(1, op, level)

    elif isinstance(root_expr, Field):
        if hasattr(root_expr, 'field_name'):
            field_name = root_expr.field_name
            ops.append(0, lambda pkt, *vargs, **kargs: getattr(pkt, field_name), level, 'field-lookup ' + repr(root_expr))
        else:
            ops.append(0, lambda pkt, *vargs, **kargs: root_expr, level, 'literal-field-value ' + repr(root_expr))
    else:
        raise Exception("Invalid argument of type %s" % repr(type(root_expr)))


    return ops

def exec_compiled_expr(pkt, args, ops, *vargs, **kargs):
    args = list(args)

    for arg_count, op in ops:
        if arg_count == 0:
            result = op(pkt, *vargs, **kargs)
        else:
            result = op(*reversed(args[:arg_count]))
            del args[:arg_count]

        args.insert(0, result)

    assert len(args) == 1
    return args[0]

def compile_expr_into_callable(root_expr):
    ops = compile_expr(root_expr).as_list()
    args = []
    return lambda pkt, *vargs, **kargs: exec_compiled_expr(pkt, args, ops, *vargs, **kargs)


def _defer_method(target, methodname, op, is_binary, is_nary=False, swap_binary_arguments=False):
    if is_binary:
        if swap_binary_arguments:
            setattr(target, methodname, lambda A, B: BinaryExpr(B, A, op))
        else:
            setattr(target, methodname, lambda A, B: BinaryExpr(A, B, op))
    else:
        if is_nary:
            def nary(A, *B, **C):
                assert B or C
                assert not (B and C)

                is_keyword_call = bool(C)

                if not C and len(B) == 1:
                    if isinstance(B[0], dict):  # nary({k1: v1, k2: v2})
                        C = B[0]
                        B = []
                    elif isinstance(B[0], (list, tuple)):   # nary([v1, v2])
                        B = B[0]
                        C = {}

                    else:
                        raise Exception("Invalid argument for nary expression '%s'. Valid arguments can be a single list or dict (like nary([a, b]) or nary({k1: a, k2: b})), a list of arguments (like nary(a, b)) or a keyword arguments call (like nary(k1=a, k2=b))." % methodname)

                elif C:                                 # nary(k1=v1, k2=v2)
                    def _encode_to_ascii_or_fail(obj):
                        if not isinstance(obj, text_types):
                            return obj # as is

                        try:
                            return obj.encode('ascii')
                        except:
                            raise Exception("Invalid argument for nary expression '%s'. Your are using a keyword arguments call (like nary(k1=a, k2=b)) where the keywords aren't valid ascii names (chars between 0 and 128)." % methodname)

                    C = {_encode_to_ascii_or_fail(k): v for k, v in C.items()}

                # else  --> nary(v1, v2)

                return NaryExpr(A, B, C, op)

            setattr(target, methodname, nary)
        else:
            setattr(target, methodname, lambda A: UnaryExpr(A, op))


def _defer_operations_of(cls, allowed_categories='all'):
    if allowed_categories == 'all':
        allowed_categories = AllCategories

    else:
        allowed_categories = list(set(allowed_categories)) # remove duplicates
        assert all((category in AllCategories) for category in allowed_categories) # sanity check

    allowed_binary_operations = sum((BinaryOperationsByCategory[category] for category in allowed_categories), [])
    allowed_binary_reverse_operations = sum((BinaryReverseOperationsByCategory[category] for category in allowed_categories), [])
    allowed_unary_operations  = sum((UnaryOperationsByCategory[category]  for category in allowed_categories), [])

    for binary_op in allowed_binary_operations:
        op_name = binary_op.__name__
        if op_name.endswith("_"):
            op_name = op_name[:-1]

        methodname = "__%s__" % op_name
        _defer_method(cls, methodname, binary_op, is_binary=True)

    for binary_op in allowed_binary_reverse_operations:
        op_name = binary_op.__name__
        if op_name.endswith("_"):
            op_name = op_name[:-1]

        methodname = "__r%s__" % op_name
        _defer_method(cls, methodname, binary_op, is_binary=True, swap_binary_arguments=True)

    for unary_op in allowed_unary_operations:
        op_name = unary_op.__name__
        if op_name.endswith("_"):
            op_name = op_name[:-1]

        if op_name == "inv":
            op_name = "invert"
        elif op_name == "truth":
            op_name = "nonzero"

        methodname = "__%s__" % op_name
        _defer_method(cls, methodname, unary_op, is_binary=False)

    _defer_method(cls, 'if_true_then_else', if_true_then_else, is_binary=False, is_nary=True)
    _defer_method(cls, 'choose', choose, is_binary=False, is_nary=True)

    return cls

def defer_operations(allowed_categories='all'):
    def decorator(cls):
        return _defer_operations_of(cls, allowed_categories)

    return decorator

UnaryExpr = defer_operations()(UnaryExpr)
BinaryExpr = defer_operations()(BinaryExpr)
NaryExpr = defer_operations()(NaryExpr)

